from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa: Counter, OrderedDict, 
from builtins import *  # noqa:

import configparser
import logging
import logging.config
import os
import json
import platform

from traceback import print_exc, format_exc

import pandas as pd
from pandas import np

from pugnlp.util import dict2obj


if platform.mac_ver()[0]:
    PLATFORM = 'OSX'
elif platform.win32_ver()[0]:
    PLATFORM = 'Win32'
elif platform.dist()[0]:
    PLATFORM = 'Linux'
else:
    PLATFORM = platform.system()  # 'Linux', 'Windows', 'Java', 'OS2', 'Darwin'

USER_HOME = os.path.expanduser("~")
PROJECT_PATH = os.path.dirname(os.path.dirname(__file__))
DATA_PATH = os.path.join(os.path.dirname(__file__), 'data')
BIGDATA_PATH = os.path.join(os.path.dirname(__file__), 'bigdata')

DEFAULT_SECRETS_PATH = os.path.join(PROJECT_PATH, 'secrets.cfg')
DEFAULT_JSON_DIR = os.path.join(DATA_PATH, 'dd')
DEFAULT_JSON_PATH = os.path.join(DEFAULT_JSON_DIR, 'bing_nodes_online', 'day_1.json')

DEFAULT_DB_DIR = os.path.join(DATA_PATH, 'db')
DEFAULT_DB_CSV_FILENAME = 'db.csv.gz'
DEFAULT_DB_CSV_PATH = os.path.join(DEFAULT_DB_DIR, DEFAULT_DB_CSV_FILENAME)
DEFAULT_ANOMALIES_CSV_FILENAME = 'anomalies.csv.gz'
DEFAULT_ANOMALIES_CSV_PATH = os.path.join(DEFAULT_DB_DIR, DEFAULT_DB_CSV_FILENAME)
DEFAULT_META_FILENAME = 'meta.json'
DEFAULT_META_PATH = os.path.join(DEFAULT_DB_DIR, DEFAULT_META_FILENAME)
DEFAULT_MODEL_FILENAME = 'model.pkl'
DEFAULT_MODEL_PATH = os.path.join(DEFAULT_DB_DIR, DEFAULT_MODEL_FILENAME)
DEFAULT_HUMAN_FILENAME = 'human-labeled-time-spans.csv'
DEFAULT_HUMAN_PATH = os.path.join(DEFAULT_DB_DIR, DEFAULT_HUMAN_FILENAME)  # start, end, 0/1 (1=anomalous)
DEFAULT_CONFIG_FILENAME = 'config.cfg'
DEFAULT_CONFIG_PATH = os.path.join(DEFAULT_DB_DIR, DEFAULT_CONFIG_FILENAME)

OSX_LOG_ADDRESS, OSX_LOG_FACILITY = os.path.join('/', 'var', 'run', 'syslog'), 'local7'
LINUX_LOG_ADDRESS, LINUX_LOG_FACILITY = os.path.join('/', 'dev', 'log'), 'local7'
LOCALHOST_LOG_ADDRESS, LOCALHOST_LOG_FACILITY = ('localhost', 514), 19

LOG_FACILITY = LINUX_LOG_FACILITY
if os.path.exists(OSX_LOG_ADDRESS):
    LOG_ADDRESS = OSX_LOG_ADDRESS
elif os.path.exists(LINUX_LOG_ADDRESS):
    LOG_ADDRESS = LINUX_LOG_ADDRESS
else:
    LOG_ADDRESS = LOCALHOST_LOG_ADDRESS
    LOG_FACILITY = LOCALHOST_LOG_FACILITY

NAME_STRIP_CHRS = '\t\n\r :-=+!._$%#@[]'

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,

    'formatters': {
        'django': {
            'format': 'django: %(message)s',
        },
        u'basic': {
            u'format': u'%(asctime)s | %(name)15s:%(lineno)3s:%(funcName)15s | %(levelname)7s | %(message)s',
        }
    },

    'handlers': {
        # 'logging.handlers.SysLogHandler': {
        #     'level': 'DEBUG',
        #     'class': 'logging.handlers.SysLogHandler',
        #     'formatter': 'django',
        #     'facility': LOG_FACILITY,
        #     'address': LOG_ADDRESS,
        # },
        'logging.handlers.SysLogHandler': {
            'level': 'DEBUG',
            'class': 'logging.handlers.SysLogHandler',
            'formatter': 'django',
            'facility': LOG_FACILITY,
            'address': LOG_ADDRESS,
        },
        u'console': {
            u'class': u'logging.StreamHandler',
            u'level': u'DEBUG',
            u'formatter': u'basic',
            u'stream': u'ext://sys.stdout',
        },
    },

    'loggers': {
        # 'loggly': {
        #     'handlers': [u'console', 'logging.handlers.SysLogHandler'],
        #     'propagate': True,
        #     'format': 'django: %(message)s',
        #     'level': 'DEBUG',
        # },
        'anom': {
            'handlers': [u'console'],
            'propagate': True,
            'level': 'DEBUG',
        },

    },
}

try:
    logging.config.dictConfig(LOGGING)
    logger = logging.getLogger(__name__)
except:
    print_exc()
    LOGGING['handlers']['logging.handlers.SysLogHandler']['address'] = LOCALHOST_LOG_ADDRESS
    LOGGING['handlers']['logging.handlers.SysLogHandler']['facility'] = 19
    try:
        logging.config.dictConfig(LOGGING)
        logger = logging.getLogger(__name__)
    except:
        print_exc()
        del LOGGING['handlers']['logging.handlers.SysLogHandler']
        LOGGING['loggers']['loggly']['handlers'] = [u'console'],
        logging.config.dictConfig(LOGGING)
        logger = logging.getLogger(__name__)


def parse_config(path=DEFAULT_CONFIG_PATH, section=None, eval_keys=['metrics', 'queries']):
    """ Pares a cfg file, retrieving the requested section as a dictionary

    Args:
        section (str): name of the section you'd like to extract
        eval_keys (list of str): Parser will try to evaluate strings in the config variables for the indicated eval_keys
    """
    configreader = configparser.RawConfigParser()
    try:
        configreader.read(path)
        configdict = configreader._sections
        configdict = configdict[section] if section else configdict
    except IOError:
        logger.error('Unable to load/parse .cfg file at "{}". Does it exist?'.format(
            path))
        configdict = {}
    for k in eval_keys:
        if k in configdict:
            try:
                # can't sandbox eval with {'__builtins__': {}} # py3.5 or {'__builtins__': None} # py2.7
                # because `float('nan')` requires builtins
                configdict[k] = eval(configdict[k], {'pd': pd, 'np': np}, {})
            except:
                logger.error('Unable to eval(configdict[{}]):\n{}'.format(k, configdict[k]))
                msg = format_exc()
                logger.error(msg)
                print(msg)
    return dict2obj(configdict)


# these should be overridden by command line args or non-default config file paths
CFG = parse_config(DEFAULT_CONFIG_PATH, section='chasedown')
SECRETS = parse_config(DEFAULT_SECRETS_PATH, section=None)
if os.path.isfile(DEFAULT_META_PATH):
    with open(DEFAULT_META_PATH) as f:
        META = json.load(f)
else:
    META = {'hosts': [], 'metrics': [], 'monitors': {}}

# str treated as *str* globstar filter
if isinstance(CFG.metrics, str):
    CFG.metrics = [s for s in META['metrics'] if CFG.metrics in s]
elif hasattr(CFG.metrics, 'match') and hasattr(CFG.metrics, 'pattern'):
    CFG.metrics = [s for s in META['metrics'] if CFG.metrics.match(s)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
`python setup.py install` install the command `chasedown` inside your current environment.

For help/docs type `chasedown -h` in your console/shell
"""
from __future__ import print_function, unicode_literals, division, absolute_import
from future import standard_library
standard_library.install_aliases()  # noqa
from builtins import *  # noqa

import argparse
import sys
import os
import datetime

import pandas as pd
from nlpia.data.loaders import read_csv

from anomalous import __version__
from anomalous.constants import logging, parse_config
from anomalous.constants import DEFAULT_DB_CSV_FILENAME, DEFAULT_DB_DIR, DEFAULT_CONFIG_FILENAME
from anomalous.utils import stdout_logging, argparse_open_file, argparse_datetime_span, clean_dd_df, update_db
from anomalous.utils import plot_predictions, update_config, clean_dd_all

__author__ = "Hobson Lane"
__copyright__ = "AuthorityLabs"
__license__ = "none"

logger = logging.getLogger(__name__)


def parse_args(args):
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace
    """
    parser = argparse.ArgumentParser(
        description="(Down)load DataDog json files into a pandas DataFrame (flatfile DB) and interactively flag anomalies")
    parser.add_argument(
        '--version',
        action='version',
        version='anomalous {ver}'.format(ver=__version__))
    parser.add_argument(
        '-l', '--load',
        dest="file_or_none", required=False,  # default=os.path.join(DATA_PATH, 'dd', 'bing_nodes_online', 'day_1.json'),
        help="Path to a DataDog json dump of server monitor time series, e.g. anomalous/data/dd/bing_nodes_online/day_1.json",
        type=lambda s: argparse_open_file(parser, s, mode='r', allow_none=True),
        metavar="FILE")
    parser.add_argument(
        '-d', '--db', '--database',
        dest="db", required=False,
        help="Path to a directory containing .csv.gz, .pkl, and .cfg files containing historical data, models, and configuration files.",
        type=str,
        metavar="DIR")
    parser.add_argument(
        '--db_cfg',
        dest="db_cfg", required=False,
        help="Path to a .cfg file containing the configuration for chasedown (timespands, hosts, etc).",
        type=str,
        metavar="CFGFILE")
    parser.add_argument(
        '--db_csv',
        dest="db_csv", required=False,
        help="Path to a .csv.gz.",
        type=str,
        metavar="CSVFILE")
    parser.add_argument(
        '-t', '--timespan',
        dest="timespan", required=False,
        help="time span to query datadog for metrics and append to a local 'database' (csv) of historical data, e.g. 'past 24 hours'",
        type=lambda s: argparse_datetime_span(parser, s, allow_none=True),
        metavar="STR")
    parser.add_argument(
        '-s', '--servers',
        help="Servers (FQDN or hostnames) for servers with Datadog monitor metrics to retrieve.",
        type=str,
        metavar="SERVERS")
    parser.add_argument(
        '-m', '--metrics',
        help="Metric names or filters on metric names.",
        type=str,
        metavar="METRICS")
    parser.add_argument(
        '-v', '--verbose',
        dest="loglevel", default=logging.WARN,
        help="Verbose stdout logging (set loglevel to INFO).",
        action='store_const',
        const=logging.INFO)
    parser.add_argument(
        '-vv', '--very-verbose',
        dest="loglevel",
        help="Very verbose logging (set loglevel to DEBUG).",
        action='store_const',
        const=logging.DEBUG)

    return parser.parse_args(args)


def main(args):
    """Main entry point allowing external calls

    Args:
      args ([str,...]): command line parameter list
    """
    args = parse_args(args)
    stdout_logging(args.loglevel)
    msg = "Arguments:\n{}".format(args.__dict__)
    logger.info(msg)
    args.db = args.db or DEFAULT_DB_DIR
    args.db_cfg = args.db_cfg or os.path.join(args.db, DEFAULT_CONFIG_FILENAME)
    cfg = parse_config(path=args.db_cfg, section='chasedown')
    cfg.timespan = argparse_datetime_span(None, getattr(cfg, 'timespan', None), allow_none=True)
    msg = "Config:\n{}".format(cfg.__dict__)
    logger.info(msg)
    cfg = update_config(cfg, args)
    cfg.db_csv = cfg.db_csv or os.path.join(cfg.db, DEFAULT_DB_CSV_FILENAME)
    msg = "cfg.update(args):\n{}".format(cfg.__dict__)
    logger.info(msg)

    # TODO: datadog query is first priority, but can't we do both a query and a file?
    if isinstance(cfg.timespan, tuple):
        start, end = cfg.timespan
    elif cfg.file_or_none is None:
        now = datetime.datetime.now()
        start, end = (now - datetime.timedelta(1), now)

    if cfg.file_or_none is None:
        df = update_db(metric_names=cfg.metrics, start=start, end=end, db=cfg.db_csv, drop=False, save=False)
    else:
        df = clean_dd_df(cfg.file_or_none)

    msg = "(Down)loaded {} series from {} with shape {}".format(
        len(df.columns), cfg.file_or_none, df.shape)
    logger.info(msg)
    logger.debug(df.describe())

    if os.path.isfile(cfg.db_csv):
        db = read_csv(cfg.db_csv)
    else:
        db = pd.DataFrame()
    db = db.append(df)

    db = clean_dd_all(db)
    plot_predictions(db.loc[datetime.datetime.now() - datetime.timedelta(1):])


def run():
    """Entry point for console_scripts
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    run()
